/* The humble pope Neige the reversing schemer and avatar of lazyness
 * officially declares this extractor and insertor for planar 2bpp tiles
 * to be (K) 3176 YOLD
 * Do what thou wilt with thine copy
*/

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#ifdef __GNUC__
#include <unistd.h>
#endif

#define OPER_EXTRACT  1
#define OPER_INSERT   2

#define TILE_LENGTH   16
#define TILE_WIDTH    8
#define TILE_HEIGTH   8

typedef uint8_t byte;
typedef uint16_t word;
typedef uint32_t dword;

#pragma pack(push, 1) /* Use 1 byte alignments */

struct color {
  byte blue;
  byte green;
  byte red;
  byte reserved;
};

/* bitmap header for 4bpp image */
struct bitmapHeader {
  /* bitmap file header */
  word signature;               /* 0x4d42 ("BM") */
  dword  fileSize;              /* sizeof(bitmapHeader) + bmpBufferLength */
  word reserved1;               /* 0 */
  word reserved2;               /* 0 */
  dword  locationOfBitmapData;    /* 0x76 */
  
  /* bitmap info header */
  dword  lengthOfHeader;          /* 40 */
  dword  imageWidth;            /* inColumns * TILE_WIDTH */
  dword  imageHeigth;           /* inRows * TILE_HEIGTH */
  word numberOfColorPlanes;     /* 1 */
  word bitPerPixel;             /* 4 */
  dword  compressionMethod;       /* 0 */
  dword  lengthOfBitmapData;    /* bmpBufferLength */
  dword  horizontalResolution;    /* 3780 */
  dword  verticalResolution;      /* 3780 */
  dword  numberOfColors;          /* 16 */
  dword  numberOfImportantColors; /* 16 */
#ifdef __GNUC__
} __attribute__ ((packed));
#else
};
#endif

#pragma pack(pop)

void oneRomTileToBitmapTile(byte *romData, byte *bmpData, int lineLength) {
  int curTileByte = 0, curBmpByte = 0;
  int line, byteInLine;
  byte pix1, pix2; /* 2 pixels to a byte in 16 colors bmp */
  byte bp1 = 0, bp2 = 0; /* bitplanes */

  for (line = 0; line < TILE_HEIGTH; line++) { /* loop the lines of the tile */
    bp1 = romData[curTileByte];
    bp2 = romData[curTileByte + 8];
    for (byteInLine = 0; byteInLine < 4; byteInLine++) { /* 4 bytes to a bmp line */
      pix1 = pix2 = 0;
      pix1 |= (bp1 & 0x80) >> 7; /* load bitplane 1 bit for pixel 1 */
      pix1 |= (bp2 & 0x80) >> 6; /* load bitplane 2 bit for pixel 1 */
      bp1 <<= 1;
      bp2 <<= 1;
      pix2 |= (bp1 & 0x80) >> 7; /* load bitplane 1 bit for pixel 2 */
      pix2 |= (bp2 & 0x80) >> 6; /* load bitplane 2 bit for pixel 2 */
      bp1 <<= 1;
      bp2 <<= 1;
      bmpData[curBmpByte + byteInLine] = (pix1 << 4) | pix2;
    }
    curTileByte++;
    curBmpByte += lineLength;
  }
}

void oneBitmapTileToRomTile(byte *romData, byte *bmpData, int lineLength) {
  int curTileByte = 0, curBmpByte = 0;
  int line, byteInLine;
  byte bp1, bp2; /* bitplanes */

  for (line = 0; line < TILE_HEIGTH; line++) {
    bp1 = bp2 = 0;
    for (byteInLine = 0; byteInLine < 4; byteInLine++) {
      bp1 <<= 1;
      bp2 <<= 1;
      bp1 |= (bmpData[curBmpByte + byteInLine] & 0x10) >> 4;
      bp2 |= (bmpData[curBmpByte + byteInLine] & 0x20) >> 5;
      bp1 <<= 1;
      bp2 <<= 1;
      bp1 |= bmpData[curBmpByte + byteInLine] & 1;
      bp2 |= (bmpData[curBmpByte + byteInLine] & 2) >> 1;
    }
    romData[curTileByte] = bp1;
    romData[curTileByte + 8] = bp2;
    curTileByte++;
    curBmpByte += lineLength;
  }
}

void extractOrInsert(byte *romData, byte *bmpData, long nmbCols, long nmbRows, long nmbTiles, char operation) {
  int curRow;
  int curCol;
  int tileOffset, bmpOffset;
  int lastTileOnLastRow = ((nmbTiles % nmbCols) ? (nmbTiles % nmbCols) : nmbCols);
  int lineLength = (((nmbCols * TILE_WIDTH * 4) + (((nmbCols * TILE_WIDTH * 4) % 32) ? (32 - ((nmbCols * TILE_WIDTH * 4) % 32)) : 0)) / 8);

  for (curRow = 0; curRow < nmbRows; curRow++) {
    for (curCol = 0; curCol < nmbCols; curCol++) {
      if ((curRow == (nmbRows - 1)) && (curCol == lastTileOnLastRow)) /* if this is the last row */
        break;
      tileOffset = (nmbCols * curRow + curCol) * TILE_LENGTH;
      bmpOffset =  (curRow * lineLength * TILE_HEIGTH) + (curCol * 4); /* 4 bytes per tile line */
      if (operation == OPER_EXTRACT)
        oneRomTileToBitmapTile(romData + tileOffset, bmpData + bmpOffset, lineLength);
      else if (operation == OPER_INSERT)
        oneBitmapTileToRomTile(romData + tileOffset, bmpData + bmpOffset, lineLength);
    }
  }
}

/* Flip image */
void flipImage(byte *bmpData, int lineLen, int numLines) {
  byte *tempBuf;
  int line;

  if ((tempBuf = (byte *)malloc(numLines * lineLen)) == NULL) {
    printf("unable to allocate %d byte for temp bitmap buffer\n", numLines * lineLen);
    exit(1);
  }
  memcpy(tempBuf, bmpData, numLines * lineLen);

  for (line = 0; line < numLines; line++)
    memcpy(bmpData + ((numLines - 1) * lineLen) - (line * lineLen), tempBuf + line * lineLen, lineLen);

  free(tempBuf);
}

#ifdef WINDOWS

/* Convert slashes to backslashes */
void unix2dos(char *fname) {
  int len = strlen(fname);
  int i;
  
  for (i = 0; i < len; i++)
    if (fname[i] == '/')
      fname[i] = '\\';
}

#endif

void usage() {
  printf("\n2x1bpp - extract or insert 2bpp planar (2 * 1bpp) 8x8 tiles from/to roms\n");
  printf("Usage: 2bpp-pc [options] -t nmbtiles romfile bmpfile\n\n");
  printf("General Options:\n");
  printf("  -t\tNumber of tiles to extract or insert\n");
  printf("  -o\tOffset in rom file to start extracting or inserting, default = 0\n");
  printf("  -e\tExtract tiles from rom\n");
  printf("  -i\tInsert tiles to rom\n\n");
  printf("Extract Options:\n");
  printf("  -c\tNumber of columns in bitmap, default = 16\n");
  printf("  -0\tBackground color in bitmap, default=0x000000 (black)\n");
  printf("  -1\tColor 1 in bitmap, default=0x666666 (dark gray)\n");
  printf("  -2\tColor 2 in bitmap, default=0xcccccc (light gray)\n");
  printf("  -3\tColor 3 in bitmap, default=0xffffff (white)\n");
  printf("Insert Options:\n");
  printf("  -n\tCreate new file, default off\n");
}

int main (int argc, char *argv[]) {
  char romFilename[FILENAME_MAX] = {0}, bmpFilename[FILENAME_MAX] = {0};
  long romFileOffset = 0;
  long inLength = 0, inColumns = 0, inRows = 0, nmbRowsToInsert = 0;
  FILE *romFile = NULL, *bmpFile = NULL;
  byte *romBuffer = NULL, *bmpBuffer = NULL;
  unsigned long romBufferLength, bmpBufferLength;
  int bmpPixPad, bmpPixWidth, bmpPixHeigth;
  char operation = OPER_EXTRACT; /* insert or extract graphics */
  struct color palette[16] = {
    {0, 0, 0, 0},
    {0x66, 0x66, 0x66, 0},
    {0xcc, 0xcc, 0xcc, 0},
    {0xff, 0xff, 0xff, 0}};
  char newFile = 0;
  long colorTemp;
  char argError = 0;
  struct bitmapHeader bmpFileHeader;
  int i;
  
  for (i = 1; i < argc; ++i)
  {
    if (strcmp(argv[i], "-t") == 0)
    {
      inLength = strtol(argv[++i], NULL, 0);
      continue;
    }
    if (strcmp(argv[i], "-c") == 0)
    {
      inColumns = strtol(argv[++i], NULL, 0);
      continue;
    }
    if (strcmp(argv[i], "-e") == 0)
    {
      operation = OPER_EXTRACT;
      continue;
    }
    if (strcmp(argv[i], "-i") == 0)
    {
      operation = OPER_INSERT;
      continue;
    }
    if (strcmp(argv[i], "-o") == 0)
    {
      romFileOffset = strtol(argv[++i], NULL, 0);
      continue;
    }
    if (strcmp(argv[i], "-0") == 0)
    {
      colorTemp = strtol(argv[++i], NULL, 0) << 8;
      palette[0].blue = (byte)(colorTemp >> 8) & 0xff;
      palette[0].green = (byte)(colorTemp >> 16) & 0xff;
      palette[0].red = (byte)(colorTemp >> 24) & 0xff;
      continue;
    }
    if (strcmp(argv[i], "-1") == 0)
    {
      colorTemp = strtol(argv[++i], NULL, 0) << 8;
      palette[1].blue = (byte)(colorTemp >> 8) & 0xff;
      palette[1].green = (byte)(colorTemp >> 16) & 0xff;
      palette[1].red = (byte)(colorTemp >> 24) & 0xff;
      continue;
    }
    if (strcmp(argv[i], "-2") == 0)
    {
      colorTemp = strtol(argv[++i], NULL, 0) << 8;
      palette[2].blue = (byte)(colorTemp >> 8) & 0xff;
      palette[2].green = (byte)(colorTemp >> 16) & 0xff;
      palette[2].red = (byte)(colorTemp >> 24) & 0xff;
      continue;
    }
    if (strcmp(argv[i], "-3") == 0)
    {
      colorTemp = strtol(argv[++i], NULL, 0) << 8;
      palette[3].blue = (byte)(colorTemp >> 8) & 0xff;
      palette[3].green = (byte)(colorTemp >> 16) & 0xff;
      palette[3].red = (byte)(colorTemp >> 24) & 0xff;
      continue;
    }
    if (strcmp(argv[i], "-n") == 0)
    {
      newFile = 1;
      continue;
    }
    /*
    if (strcmp(argv[i], "-") == 0)
    {
      continue;
    }
    */
    if (argv[i][0] != '-')
    {
      if (romFilename[0] == '\0') {
        strncpy(romFilename, argv[i], FILENAME_MAX);
        continue;
      }
      else if (bmpFilename[0] == '\0') {
        strncpy(bmpFilename, argv[i], FILENAME_MAX);
        continue;
      }        
    }

    printf("invalid option: %s\n", argv[i]);
    usage();
    exit(1);
  }
  
  if (romFilename[0] == '\0') {
    printf("missing rom filename\n");
    argError++;
  }
  
  if (bmpFilename[0] == '\0') {
    printf("missing bitmap filename\n");
    argError++;
  }
  
  if (!inLength) {
    printf("missing number of tiles to extract/insert\n");
    argError++;
  }
  
  if (argError) {
    usage();
    exit(1);
  }

  if (!inColumns)
    inColumns = 16;

#ifdef WINDOWS
  unix2dos(romFilename);
  unix2dos(bmpFilename);
#endif

  if (operation == OPER_EXTRACT) {
    /*
     * each line of bitmap data is aligned on 32 bits.
     *
     * get the number of pixels (bits) in the image and divide by 8
     * to get the number of bytes to allocate buffer (possibility of overflow)
     *
     * if number of bit in line is not equally divisible by 32,
     * you need to add the missing bits to align the data (32 - missing_bits).
     *
     * padding = ((col * TILE_WIDTH * BPP) % 32) ? 32 - ((col * TILE_WIDTH * BPP) % 32) : 0
     * xlength = (col * TILE_WIDTH * BPP) + padding
     * ylength = row * TILE_HEIGTH
     * bufflen = xlength * ylength / 8
     */
     
    inRows = (inLength / inColumns) + ((inLength % inColumns) ? 1 : 0);
    romBufferLength = inLength * TILE_LENGTH;
    bmpPixPad = ((inColumns * TILE_WIDTH * 4) % 32) ? 32 - ((inColumns * TILE_WIDTH * 4) % 32) : 0;
    bmpPixWidth = (inColumns * TILE_WIDTH * 4) + bmpPixPad;
    bmpPixHeigth = inRows * TILE_HEIGTH;
    bmpBufferLength = bmpPixWidth * bmpPixHeigth / 8;
  
    if ((romFile = fopen(romFilename, "rb")) == NULL) {
      printf("unable to open rom file %s for reading\n", romFilename);
      exit(1);
    }
    
    if ((bmpFile = fopen(bmpFilename, "wb")) == NULL) {
      printf("unable to open bitmap file %s for writing\n", bmpFilename);
      exit(1);
    }
    
    if ((romBuffer = (byte *)malloc(romBufferLength)) == NULL) {
      printf("unable to allocate %ld byte for rom buffer\n", romBufferLength);
      exit(1);
    }

    if ((bmpBuffer = (byte *)malloc(bmpBufferLength)) == NULL) {
      printf("unable to allocate %ld byte for bitmap buffer\n", bmpBufferLength);
      exit(1);
    }
    memset(bmpBuffer, 0, bmpBufferLength);
    
    fseek(romFile, romFileOffset, SEEK_SET);
    
    if (fread(romBuffer, 1, romBufferLength, romFile) < romBufferLength) {
      printf("%s rom file is too small", romFilename);
      exit(1);
    }

    extractOrInsert(romBuffer, bmpBuffer, inColumns, inRows, inLength, OPER_EXTRACT);

    flipImage(bmpBuffer, bmpPixWidth / 8, bmpPixHeigth);

    bmpFileHeader.signature = 0x4d42;          /* 0x4d42 ("BM") */
    bmpFileHeader.fileSize = sizeof(bmpFileHeader) + bmpBufferLength; /* sizeof(bitmapHeader) + sizeof(bmpFileBuffer) */
    bmpFileHeader.reserved1 = 0;               /* 0 */
    bmpFileHeader.reserved2 = 0;               /* 0 */
    bmpFileHeader.locationOfBitmapData = 0x76; /* 0x76 */
  
    /* bitmap info header */
    bmpFileHeader.lengthOfHeader = 40;         /* 40 */
    bmpFileHeader.imageWidth = inColumns * TILE_WIDTH; /* inColumns * TILE_WIDTH */
    bmpFileHeader.imageHeigth = inRows * TILE_HEIGTH;   /* inRows * TILE_HEIGTH */
    bmpFileHeader.numberOfColorPlanes = 1;     /* 1 */
    bmpFileHeader.bitPerPixel = 4;             /* 4 */
    bmpFileHeader.compressionMethod = 0;       /* 0 */
    bmpFileHeader.lengthOfBitmapData = bmpBufferLength;    /* bmpBufferLength */
    bmpFileHeader.horizontalResolution = 3780; /* 3780 */
    bmpFileHeader.verticalResolution = 3780;   /* 3780 */
    bmpFileHeader.numberOfColors = 16;          /* 0 */
    bmpFileHeader.numberOfImportantColors = 16; /* 0 */

    fwrite(&bmpFileHeader, sizeof(bmpFileHeader), 1, bmpFile);
    fwrite(palette, sizeof(struct color), 16, bmpFile);
    fwrite(bmpBuffer, 1, bmpBufferLength, bmpFile);
  }
  else if (operation == OPER_INSERT) {
    if (newFile) {
      romFile = fopen(romFilename, "wb");
    }
    else {
      romFile = fopen(romFilename, "rb+");
    }
    if (romFile == NULL) {
      printf("unable to open rom file %s for writing\n", romFilename);
      exit(1);
    }
    
    if ((bmpFile = fopen(bmpFilename, "rb")) == NULL) {
      printf("unable to open bitmap file %s for reading\n", bmpFilename);
      exit(1);
    }
    
    if (fread(&bmpFileHeader, sizeof(bmpFileHeader), 1, bmpFile) < 1) {
      printf("bitmap file %s is too small\n", bmpFilename);
      exit(1);
    }
    
    if (bmpFileHeader.signature != 0x4d42) {
      printf("%s is not a bitmap file\n", bmpFilename);
      exit(1);
    }
    
    if (bmpFileHeader.bitPerPixel != 4) {
      printf("%s is not a 16 color bitmap\n", bmpFilename);
      exit(1);
    }
    
    inColumns = bmpFileHeader.imageWidth / TILE_WIDTH;
    inRows = bmpFileHeader.imageHeigth / TILE_HEIGTH;
    bmpBufferLength = bmpFileHeader.lengthOfBitmapData;
    romBufferLength = inLength * TILE_LENGTH;
    nmbRowsToInsert = inLength / inColumns;
    nmbRowsToInsert = ((inLength % inColumns) ? nmbRowsToInsert + 1 : nmbRowsToInsert);
   
    if ((inColumns * inRows) < inLength) {
      printf("bitmap file %s is too small\n", bmpFilename);
      exit(1);
    }
  
    if ((romBuffer = (byte *)malloc(romBufferLength)) == NULL) {
      printf("unable to allocate %ld byte for rom buffer\n", romBufferLength);
      exit(1);
    }
    memset(romBuffer, 0, romBufferLength);

    if ((bmpBuffer = (byte *)malloc(bmpBufferLength)) == NULL) {
      printf("unable to allocate %ld byte for bitmap buffer\n", bmpBufferLength);
      exit(1);
    }

    fseek(bmpFile, bmpFileHeader.locationOfBitmapData, SEEK_SET);
    if (fread(bmpBuffer, 1, bmpBufferLength, bmpFile) < bmpBufferLength) {
      printf("%s bitmap file is too small", bmpFilename);
      exit(1);
    }

    flipImage(bmpBuffer, bmpFileHeader.imageWidth / 2, bmpFileHeader.imageHeigth);

    extractOrInsert(romBuffer, bmpBuffer, inColumns, nmbRowsToInsert, inLength, OPER_INSERT);
    
    if (romFileOffset > 0)
      fseek(romFile, romFileOffset, SEEK_SET);
    fwrite(romBuffer, 1, romBufferLength, romFile);
  }
  
  free(romBuffer);
  free(bmpBuffer);
  fclose(romFile);
  fclose(bmpFile);
  
#ifdef DEBUG
  printf("romFilename: %s\n", romFilename);
  printf("bmpFilename: %s\n", bmpFilename);
  printf("romFileOffset: %ld\n", romFileOffset);
  printf("inLength: %ld\n", inLength);
  printf("inColumns: %ld\n", inColumns);
  printf("inRows: %ld\n", inRows);
  printf("romBufferLength: %ld\n", romBufferLength);
  printf("bmpBufferLength: %ld\n", bmpBufferLength);
  printf("operation: %s\n", ((operation < 2) ? "extract" : "insert"));
#endif

  return 0;
}
