; The humble pope Neige the reversing schemer and avatar of lazyness
; officially declares this compressor & decompressor for Dragon Ball Z - Super Gokuuden Totsugeki Hen (J)
; to be (K) 3176 YOLD
; Do what thou wilt with thine copy

#lang scheme

(require scheme/port
         "libs/io.ss")

;; bitwise-set-bit : integer? integer? -> integer?
(define (bitwise-set-bit ei1 ei2)
  (bitwise-ior (arithmetic-shift 1 ei2) ei1))

(define MAXIMUM-LENGTH 18) ; constant
(define MINIMUM-LENGTH 3)  ; constant

;; decompress : input-port? -> bytes
(define (decompress ip)
  (let* ([size (read-word ip)]
         [out (make-bytes size #xFF)]
         [read-delta-count-pair
          (λ ()
            (let ([word (read-word ip)])
              (values (bitwise-and (arithmetic-shift word -4) #x0FFF)
                      (+ (bitwise-and word #x000F) MINIMUM-LENGTH))))]
         [copy-out-part
          (λ (si di c)
            (let copy-loop ([si si] [di di] [c c])
              (unless (zero? c)
                (if (< si 0)
                    (bytes-set! out di 0)
                    (bytes-set! out di (bytes-ref out si)))
                (copy-loop (add1 si) (add1 di) (sub1 c)))))])
    (let decomp-loop ([num-uncompressed 0]
                      [out-index 0]
                      [bit-to-test 8]
                      [control-byte 0]
                      [compressed-size 0])
      (cond
        [(= num-uncompressed size) ; reached the uncompressed length, we are finished
         (printf "compressed size: ~a\n" compressed-size)
         out]
        [(> num-uncompressed size) ; length does not agree with uncompressed length
         (error 'uncompress "bad compressed data")]
        [(> bit-to-test 7) ; read control byte
         (decomp-loop num-uncompressed
                      out-index
                      0
                      (read-byte ip)
                      (add1 compressed-size))]
        [else
         (if (bitwise-bit-set? control-byte bit-to-test)
             (let ([out-byte (read-byte ip)]) ; output literal byte
               (bytes-set! out out-index out-byte)
               (decomp-loop (add1 num-uncompressed)
                            (add1 out-index)
                            (add1 bit-to-test)
                            control-byte
                            (add1 compressed-size)))
             (let-values ([(delta count) (read-delta-count-pair)]) ; read delta / count pair
               (copy-out-part (- out-index delta) out-index count)
               (decomp-loop (+ num-uncompressed count)
                            (+ out-index count)
                            (add1 bit-to-test)
                            control-byte
                            (+ compressed-size 2))))]))))

;==============================================================================

;; compress : bytes? output-port? -> natural-number/c
(define (compress data op)
  (let* ([dict-indices (make-hash)]
         [uncompressed-size (bytes-length data)]
         [make-delta-count-byte-pair
          (λ (delta count)
            (let ([word (bitwise-ior (arithmetic-shift (bitwise-and delta #x0FFF) 4)
                                     (bitwise-and (- count MINIMUM-LENGTH) #x0F))])
              (list (bitwise-and (arithmetic-shift word -8) #xFF)
                    (bitwise-and word #xFF))))]
         [repeat-index-cmd
          (λ (cmd i n)
            (for ([i (in-range i (+ i n))])
              (cmd (subbytes data i (+ i MINIMUM-LENGTH)) i)))]
         [add-index
          (λ (k i)
            (let ([exists? (member i (hash-ref dict-indices k '()))])
              (unless exists?
                (hash-set! dict-indices k (cons i (hash-ref dict-indices k '()))))))]
         [remove-index
          (λ (k i)
            (let ([new (remove* (list i) (hash-ref dict-indices k '()))])
              (if (null? new)
                  (hash-remove! dict-indices k)
                  (hash-set! dict-indices k new))))]
         [add-indices
          (λ (i n)
            (repeat-index-cmd add-index i n))]
         [remove-indices
          (λ (i n)
            (repeat-index-cmd remove-index i n))]
         [check-sequence
          (λ (i1 i2)
            (let check-loop ([i1 i1] [i2 i2] [len 0])
              (if (or (= uncompressed-size i1)
                      (= uncompressed-size i2)
                      (not (= (bytes-ref data i1) (bytes-ref data i2)))
                      (= len MAXIMUM-LENGTH))
                  len
                  (check-loop (add1 i1)
                              (add1 i2)
                              (add1 len)))))]
         [num-leading-zeroes
          (let loop ([len 0])
            (if (or (= len MAXIMUM-LENGTH)
                    (not (zero? (bytes-ref data len))))
                len
                (loop (add1 len))))]
         [find-best-sequence
          (λ (start-index)
            (letrec ([longest-sequence
                      (λ ()
                        (let ls-loop ([indices-to-check (hash-ref dict-indices
                                                (subbytes data start-index (+ start-index MINIMUM-LENGTH))
                                                '())]
                                      [longest-idx 0]
                                      [longest-len 0])
                          (if (null? indices-to-check)
                              (if (< longest-len MINIMUM-LENGTH)
                                  (values 0 0)
                                  (values longest-idx longest-len))
                              (let ([found-len (check-sequence start-index (car indices-to-check))])
                                (if (= found-len MAXIMUM-LENGTH)
                                    (values (car indices-to-check) found-len)
                                    (ls-loop (cdr indices-to-check)
                                             (if (> found-len longest-len) (car indices-to-check) longest-idx)
                                             (if (> found-len longest-len) found-len longest-len)))))))])
              (cond
                [(and (= start-index 0)
                      (< MINIMUM-LENGTH num-leading-zeroes))
                 (cons (- num-leading-zeroes) num-leading-zeroes)] ; special case
                [(> start-index (- uncompressed-size MINIMUM-LENGTH))
                 (cons #f 1)] ; not enough bytes left for sequence, return literal
                [else
                 (let-values ([(idx len) (longest-sequence)])
                   (if (< len MINIMUM-LENGTH)
                       (cons #f 1) ; sequence too short, return literal
                       (cons idx len)))])))]
         ; build the best-matches list
         [best-matches (let bm-loop ([src-index 0])
                         (when (> src-index #xFFF)
                           (let* ([delta (- src-index #xFFF 1)]
                                  [start (if (< delta 0) 0 delta)]
                                  [count (if (< delta 0) (+ delta 1) 1)])
                             (remove-indices start count))) ; remove out of range indices
                         (when (< 0 src-index (- uncompressed-size MINIMUM-LENGTH))
                           (add-indices (- src-index 1) 1)) ; add indices that entered range
                         (if (= src-index uncompressed-size)
                             '()
                             (cons (find-best-sequence src-index)
                                   (bm-loop (add1 src-index)))))]
         ; for each byte starting from the end, compute the longest match resulting in the smallest output size
         [optimal-parse (let op-loop ([ml (reverse best-matches)]
                                      [cml '()]
                                      [out '(0)])
                          (if (null? ml)
                              cml
                              (let ([chosen (let cl-loop ([n 1]
                                                          [l +inf.0]
                                                          [min-out-size +inf.0]
                                                          [out-copy out])
                                              (cond
                                                [(> n (cdar ml))
                                                 (cons l min-out-size)]
                                                [(< 1 n MINIMUM-LENGTH) ; can't encode counts between 1 and MINIMUM-LENGTH
                                                 (cl-loop (add1 n) l min-out-size (cdr out-copy))]
                                                [else
                                                 (let ([out-size (+ (if (= n 1) 1 2) (car out-copy))])
                                                   (cl-loop (add1 n)
                                                            (if (<= out-size min-out-size) n l)
                                                            (if (<= out-size min-out-size) out-size min-out-size)
                                                            (cdr out-copy)))]))])
                                (op-loop (cdr ml)
                                         (cons (cons (caar ml) (car chosen)) cml)
                                         (cons (cdr chosen) out)))))])
    (write-word uncompressed-size op) ; write the uncompressed length
    (let comp-loop ([opt optimal-parse]
                    [src-index 0]
                    [bit-to-set 0]
                    [control-byte 0]
                    [temp '()])
      (cond
        [(null? opt) ; reached end of input, we are finished
         (write-bytes (list->bytes (cons control-byte (reverse temp))) op)]
        [(> bit-to-set 7) ; filled control byte, write temp buffer
         (write-bytes (list->bytes (cons control-byte (reverse temp))) op)
         (comp-loop opt
                    src-index
                    0
                    0
                    '())]
        [else
         (let ([index (caar opt)]
               [count (cdar opt)])
           (if (> count 1) ; found string in dictionary
               (comp-loop (list-tail opt count)
                          (+ src-index count) ; output delta/count pair
                          (add1 bit-to-set)
                          control-byte
                          (append (make-delta-count-byte-pair (- src-index index) count) temp))
               (comp-loop (cdr opt)
                          (add1 src-index) ; output literal byte
                          (add1 bit-to-set)
                          (bitwise-set-bit control-byte bit-to-set)
                          (cons (bytes-ref data src-index) temp))))]))))

;==============================================================================

(define (decompress-to-file in-file out-file [in-offset 0])
  (call/if in-file
    (λ (ip)
      (unless (zero? in-offset)
        (file-position ip in-offset))
      (call/of out-file
        (λ (op)
          (write-bytes (decompress ip) op))
        #:mode 'binary #:exists 'truncate))
    #:mode 'binary))

;; read-all-bytes : path-string? -> bytes
(define (read-all-bytes in-file)
  (let ([obp (open-output-bytes)])
    (call/if in-file
      (λ (ip)
        (copy-port ip obp)
        (get-output-bytes obp))
      #:mode 'binary)))

(define (compress-to-file in-file out-file [out-offset 0])
  (let ([data (read-all-bytes in-file)])
    (call/of out-file
      (λ (op)
        (unless (zero? out-offset)
          (file-position op out-offset))
        (compress data op))
      #:mode 'binary #:exists (if (zero? out-offset ) 'truncate 'can-update))))

;(decompress-to-file "dbz-sgth-orig.smc" "decompressed.bin" #x1b9eb1)
;(compress-to-file "decompressed.bin" "compressed.bin")
;(decompress-to-file "compressed.bin" "redecompressed.bin")

;==============================================================================

(define offset (make-parameter 0))
(define operation (make-parameter 'decompress))

(define-values (rom-file-name dec-file-name)
  (command-line #:program "codec"
                #:once-each
                [("-o" "--offset") ofs
                                     "offset in the rom file to compress/decompress"
                                     (offset (string->number ofs))]
                #:once-any
                [("-c" "--compress") "compress data from dec-file to rom-file"
                                     (operation 'compress)]
                [("-d" "--decompress") "decompress data from rom-file to dec-file"
                                       (operation 'decompress)]
                #:args (rom-file dec-file)
                (values rom-file dec-file)))

#;(printf "op ~a, o #x~x, rom ~a, dec ~a\n" (operation) (offset) rom-file-name dec-file-name)

(if (eq? (operation) 'decompress)
    (decompress-to-file rom-file-name dec-file-name (offset))
    (compress-to-file dec-file-name rom-file-name (offset)))
(void)
