; The humble pope Neige the reversing schemer and avatar of lazyness
; officially declares this io library
; to be (K) 3176 YOLD
; Do what thou wilt with thine copy

#lang scheme/base

(require scheme/contract)

(provide/contract [io-big-endian? (-> boolean?)]
                  [set-io-big-endian! (-> boolean? any)]
                  [io-at (-> port? natural-number/c (-> port? any) any)]
                  [read-byte-at (-> input-port? natural-number/c (or/c eof-object? byte?))]
                  [write-byte-at (-> byte? output-port? natural-number/c any)]
                  [read-n-bytes (->* (natural-number/c input-port?)
                                     (boolean?)
                                     (or/c eof-object? natural-number/c))]
                  [write-n-bytes (->* (natural-number/c natural-number/c output-port?)
                                      (boolean?)
                                      any)]
                  [read-n-bytes-at (->* (natural-number/c input-port? natural-number/c)
                                        (boolean?)
                                        (or/c eof-object? natural-number/c))]
                  [write-n-bytes-at (->* (natural-number/c natural-number/c output-port? natural-number/c)
                                         (boolean?)
                                         any)]
                  [read-word (->* (input-port?)
                                  (boolean?)
                                  (or/c eof-object? (integer-in 0 #xFFFF)))]
                  [write-word (->* (natural-number/c output-port?)
                                   (boolean?)
                                   any)]
                  [read-word-at (->* (input-port? natural-number/c)
                                     (boolean?)
                                     (or/c eof-object? (integer-in 0 #xFFFF)))]
                  [write-word-at (->* (natural-number/c output-port? natural-number/c)
                                      (boolean?)
                                      any)]
                  [read-3bytes (->* (input-port?)
                                    (boolean?)
                                    (or/c eof-object? (integer-in 0 #xFFFFFF)))]
                  [write-3bytes (->* (natural-number/c output-port?)
                                     (boolean?)
                                     any)]
                  [read-3bytes-at (->* (input-port? natural-number/c)
                                       (boolean?)
                                       (or/c eof-object? (integer-in 0 #xFFFFFF)))]
                  [write-3bytes-at (->* (natural-number/c output-port? natural-number/c)
                                        (boolean?)
                                        any)]
                  [read-dword (->* (input-port?)
                                   (boolean?)
                                   (or/c eof-object? (integer-in 0 #xFFFFFFFF)))]
                  [write-dword (->* (natural-number/c output-port?)
                                    (boolean?)
                                    any)]
                  [read-dword-at (->* (input-port? natural-number/c)
                                      (boolean?)
                                      (or/c eof-object? (integer-in 0 #xFFFFFFFF)))]
                  [write-dword-at (->* (natural-number/c output-port? natural-number/c)
                                       (boolean?)
                                       any)]
                  [read-byte-terminated-bytes (-> byte? input-port? bytes?)]
                  [write-byte-terminated-bytes (-> bytes? byte? output-port? any)]
                  [read-byte-terminated-bytes-at (-> byte? input-port? natural-number/c bytes?)]
                  [write-byte-terminated-bytes-at (-> bytes? byte? output-port? natural-number/c any)]
                  [read-char-terminated-string (-> char? input-port? string?)]
                  [write-char-terminated-string (-> string? char? output-port? any)]
                  [read-char-terminated-string-at (-> char? input-port? natural-number/c string?)]
                  [write-char-terminated-string-at (-> string? char? output-port? natural-number/c any)]
                  [read-bytesz (-> input-port? bytes?)]
                  [write-bytesz (-> bytes? output-port? any)]
                  [read-bytesz-at (-> input-port? natural-number/c bytes?)]
                  [write-bytesz-at (-> bytes? output-port? natural-number/c any)]
                  [read-stringz (-> input-port? string?)]
                  [write-stringz (-> string? output-port? any)]
                  [read-stringz-at (-> input-port? natural-number/c string?)]
                  [write-stringz-at (-> string? output-port? natural-number/c any)]
                  [read-pascal-bytes (->* (input-port?)
                                          ((-> input-port? any))
                                          bytes?)]
                  [write-pascal-bytes (->* (bytes? output-port?)
                                           ((-> integer? output-port? any))
                                           any)]
                  [read-pascal-bytes-at (->* (input-port? natural-number/c)
                                             ((-> input-port? any))
                                             bytes?)]
                  [write-pascal-bytes-at (->* (bytes? output-port? natural-number/c)
                                              ((-> integer? output-port? any))
                                              any)]
                  [read-pascal-string (->* (input-port?)
                                           ((-> input-port? any))
                                           string?)]
                  [write-pascal-string (->* (string? output-port?)
                                            ((-> integer? output-port? any))
                                            any)]
                  [read-pascal-string-at (->* (input-port? natural-number/c)
                                              ((-> input-port? any))
                                              string?)]
                  [write-pascal-string-at (->* (string? output-port? natural-number/c)
                                               ((-> integer? output-port? any))
                                               any)])

;; shortcuts
(provide (rename-out [call-with-input-file call/if]
                     [call-with-output-file call/of]))

;; io-big-endian? :  -> boolean
;; return #t if integer io uses the big-endian byte ordering,
;; #f if it use the little-endian byte ordering
;; set-io-big-endian! : boolean -> void
;; sets the default byte ordering to use for integer io
(define-values (io-big-endian?
                set-io-big-endian!)
  (let ([big-endian? (system-big-endian?)])
    (values
     (lambda ()
       big-endian?)
     (lambda (new-big-endian)
       (set! big-endian? new-big-endian)))))

;; io-at : port integer proc -> any
;; set the port position, execute proc and return to the original position
(define (io-at p ofs proc)
  (let ([old-ofs (file-position p)]
        [dummy (file-position p ofs)]
        [ret (proc p)])
    (file-position p old-ofs)
    ret))

;; read-byte-at : input-port integer -> integer
;; read a byte from offset ofs of input port ip
(define (read-byte-at ip ofs)
  (io-at ip ofs read-byte))

;; write-byte-at : byte output-port integer -> any
;; write byte val to offset ofs of output-port op
(define (write-byte-at val op ofs)
  (io-at op ofs (λ (op) (write-byte val op))))

#;
(define (print-bytes b)
  (let loop ([n 0])
    (unless (= n (bytes-length b))
      (printf "~a~x " (if (< (bytes-ref b n) #x10) #"0" #"") (bytes-ref b n))
      (loop (add1 n))))
  (newline))

;; read-n-bytes : integer input-port [bool] -> integer
;; read an integer of n bytes from input port ip
(define (read-n-bytes n ip [big-endian? (io-big-endian?)])
  (let ([end (if big-endian? -1 n)]
        [next (if big-endian? sub1 add1)])
    (let/ec break
      (let loop ([i (if big-endian? (sub1 n) 0)]
                 [val 0])
        (if (= i end)
            val
            (let ([byte (read-byte ip)])
              (if (eof-object? byte)
                  (break byte)
                  (loop (next i)
                        (bitwise-ior (arithmetic-shift byte (* i 8)) val)))))))))
;(printf "#x~x\n" (read-n-bytes 7 (open-input-bytes #"\xDE\xBC\x9A\x78\x56\x34\x12") #f))
;(printf "#x~x\n" (read-n-bytes 7 (open-input-bytes #"\xDE\xBC\x9A\x78\x56\x34\x12") #t))

;; write-n-bytes : integer integer output-port [bool] -> any
;; write n bytes of the integer val to output-port op
(define (write-n-bytes val n op [big-endian? (io-big-endian?)])
  (let* ([end (if big-endian? -1 n)]
         [next (if big-endian? sub1 add1)])
    (let loop ([i (if big-endian? (sub1 n) 0)])
      (unless (= i end)
        (write-byte (bitwise-and (arithmetic-shift val (- (* i 8))) #xFF) op)
        (loop (next i))))))
#;
(let ([osp (open-output-bytes)])
  (write-n-bytes #x123456789ABCDE 7 osp #f)
  (print-bytes (get-output-bytes osp 'reset))
  (write-n-bytes #x123456789ABCDE 7 osp #t)
  (print-bytes (get-output-bytes osp)))

;; read-n-bytes-at : integer input-port integer [bool] -> integer
;; read an integer of n bytes from offset ofs of input port ip
(define (read-n-bytes-at n ip ofs [big-endian? (io-big-endian?)])
  (io-at ip ofs (λ (ip) (read-n-bytes n ip big-endian?))))

;; write-n-bytes-at : output-port integer integer integer [bool] -> any
;; write n bytes of the integer val to offset ofs of output-port op
(define (write-n-bytes-at op n val ofs [big-endian? (io-big-endian?)])
  (io-at op ofs (λ (op) (write-n-bytes val n op big-endian?))))

;; read-word : input-port [bool] -> integer
;; read a word from input port ip
(define (read-word ip [big-endian? (io-big-endian?)])
  (read-n-bytes 2 ip big-endian?))

;; write-word : integer output-port [bool] -> any
;; write the word val to output port op
(define (write-word val op [big-endian? (io-big-endian?)])
  (write-n-bytes val 2 op big-endian?))

;; read-word-at : input-port integer [bool] -> integer
;; read a word from offset ofs of input port ip
(define (read-word-at ip ofs [big-endian? (io-big-endian?)])
  (io-at ip ofs (lambda (ip) (read-word ip big-endian?))))

;; write-word-at : integer output-port integer [bool] -> any
;; write the word val to offset ofs of output port op
(define (write-word-at val op ofs [big-endian? (io-big-endian?)])
  (io-at op ofs (lambda (op) (write-word val op big-endian?))))

;; read-3byte : input-port [bool] -> integer
;; read a 24-bits word from input port ip
(define (read-3bytes ip [big-endian? (io-big-endian?)])
  (read-n-bytes 3 ip big-endian?))

;; write-3byte : integer output-port [bool] -> any
;; write the 24-bits word val to output port op
(define (write-3bytes val op [big-endian? (io-big-endian?)])
  (write-n-bytes val 3 op big-endian?))

;; read-3byte-at : input-port integer [bool] -> integer
;; read a 24-bits word from offset ofs of input port ip
(define (read-3bytes-at ip ofs [big-endian? (io-big-endian?)])
  (io-at ip ofs (lambda (ip) (read-3bytes ip big-endian?))))

;; write-3byte-at : integer output-port integer [bool] -> any
;; write the 24-bits word val to offset ofs of output port op
(define (write-3bytes-at val op ofs [big-endian? (io-big-endian?)])
  (io-at op ofs (lambda (op) (write-3bytes val op big-endian?))))

;; read-dword : input-port [bool] -> integer
;; read a double word from input port ip
(define (read-dword ip [big-endian? (io-big-endian?)])
  (read-n-bytes 4 ip big-endian?))

;; write-dword : output-port val [bool] -> any
;; write the word val to output port op
(define (write-dword val op [big-endian? (io-big-endian?)])
  (write-n-bytes val 4 op big-endian?))

;; read-dword-at : input-port integer [bool] -> integer
;; read a dword from offset ofs of input port ip
(define (read-dword-at ip ofs [big-endian? (io-big-endian?)])
  (io-at ip ofs (lambda (ip) (read-dword ip big-endian?))))

;; write-dword-at : integer output-port integer [bool] -> any
;; write the word val to offset ofs of output port op
(define (write-dword-at val op ofs [big-endian? (io-big-endian?)])
  (io-at op ofs (lambda (op) (write-dword val op big-endian?))))

;; read-byte-terminated-bytes : byte input-port -> bytes
;; read a byte string from input port ip until it encounters a byte with value stop
(define (read-byte-terminated-bytes stop ip)
  (let loop ([byte (read-byte ip)])
    (if (= byte stop)
        #"" ; do not include ending byte
        (bytes-append (bytes byte) (loop (read-byte ip))))))

;; write-byte-terminated-bytes : bytes byte output-port -> any
;; write the bytes val followed by byte stop to output port op
(define (write-byte-terminated-bytes val stop op)
  (write-bytes val op)
  (write-byte stop op))
#;
(let ([osp (open-output-bytes)])
  (write-byte-terminated-bytes #"Neige" #x00 osp)
  (get-output-bytes osp))

;; read-byte-terminated-bytes-at : byte input-port integer -> bytes
;; read a byte string from offset ofs of input port ip until it encounters a byte with value stop
(define (read-byte-terminated-bytes-at stop ip ofs)
  (io-at ip ofs (λ (ip) (read-byte-terminated-bytes stop ip))))

;; write-byte-terminated-bytes-at : bytes byte output-port integer -> any
;; write the bytes val followed by byte stop to offset ofs of output port op
(define (write-byte-terminated-bytes-at val stop op ofs)
  (io-at op ofs (λ (op) (write-byte-terminated-bytes val stop op))))

;; read-char-terminated-string : char input-port -> string
;; read a string from input port ip until it encounters a char with value stop
(define (read-char-terminated-string stop ip)
  (let loop ([char (read-char ip)])
    (if (char=? char stop)
        "" ; do not include ending char
        (string-append (string char) (loop (read-char ip))))))

;; write-char-terminated-string : string char output-port -> any
;; write the string val followed by char stop to output port op
(define (write-char-terminated-string val stop op)
  (write-string val op)
  (write-char stop op))
#;
(let ([osp (open-output-bytes)])
  (write-char-terminated-string "Neige" #\nul osp)
  (get-output-bytes osp))

;; read-char-terminated-string-at : char input-port integer -> string
;; read a string from offset ofs of input port ip until it encounters a char with value stop
(define (read-char-terminated-string-at stop ip ofs)
  (io-at ip ofs (λ (ip) (read-char-terminated-string stop ip))))

;; write-char-terminated-string-at : string char output-port integer -> any
;; write the string val followed by char stop to offset of output port op
(define (write-char-terminated-string-at val stop op ofs)
  (io-at op ofs (λ (op) (write-char-terminated-string val stop op))))

;; read-bytesz : input-port -> bytes
;; read a zero terminated bytes string from input port ip
(define (read-bytesz ip)
  (read-byte-terminated-bytes 0 ip))
;(printf "~a\n" (read-bytesz (open-input-bytes #"Neige\x00ABCDEF")))

;; write-bytesz : bytes output-port -> any
;; write the zero terminated bytes string val to output port op
(define (write-bytesz val op)
  (write-byte-terminated-bytes val 0 op))
#;
(let ([osp (open-output-bytes)])
  (write-bytesz #"Neige" osp)
  (get-output-bytes osp))

;; read-bytesz-at : input-port integer -> bytes
;; read a zero terminated bytes string from offset ofs of input port ip
(define (read-bytesz-at ip ofs)
  (io-at ip ofs read-bytesz))

;; write-bytesz-at : bytes output-port integer -> any
;; write the zero terminated bytes string val to offset ofs of output port op
(define (write-bytesz-at val op ofs)
  (io-at op ofs (λ (op) (write-bytesz val op))))

;; read-stringz : input-port -> string
;; read a zero terminated string from input port ip
(define (read-stringz ip)
  (read-char-terminated-string #\nul ip))
;(printf "~a\n" (read-stringz (open-input-string "Neige\x00ABCDEF")))

;; write-stringz : string output-port -> any
;; write the zero terminated string string val to output port op
(define (write-stringz val op)
  (write-char-terminated-string val #\nul op))
#;
(let ([osp (open-output-bytes)])
  (write-stringz "Neige" osp)
  (get-output-bytes osp))

;; read-stringz-at : input-port integer -> string
;; read a zero terminated string from offset ofs of input port ip
(define (read-stringz-at ip ofs)
  (io-at ip ofs read-stringz))

;; write-stringz-at : string output-port integer -> any
;; write the zero terminated string string val to offset ofs of output port op
(define (write-stringz-at val op ofs)
  (io-at op ofs (λ (op) (write-stringz val op))))

;; read-pascal-bytes : input-port [procedure] -> bytes
;; read the length using proc then read length bytes from input port ip
(define (read-pascal-bytes ip [proc read-byte])
  (read-bytes (proc ip) ip))
;(printf "~a\n" (read-pascal-bytes (open-input-string "\x05NeigeABCDEF")))

;; write-pascal-bytes : output-port bytes [procedure] -> any
;; write the length of the bytes val using proc then write the bytes val to output port op
(define (write-pascal-bytes val op [proc write-byte])
  (proc (bytes-length val) op)
  (write-bytes val op))
#;
(let ([osp (open-output-bytes)])
  (write-pascal-bytes #"Neige" osp)
  (get-output-bytes osp))

;; read-pascal-bytes-at : input-port integer [procedure] -> bytes
;; read the length using proc then read length bytes from offset ofs of input port ip
(define (read-pascal-bytes-at ip ofs [proc read-byte])
  (io-at ip ofs (λ (ip) (read-pascal-bytes ip proc))))

;; write-pascal-bytes-at : output-port bytes integer [procedure] -> any
;; write the length of the bytes val using proc then write the bytes val to offset ofs of output port op
(define (write-pascal-bytes-at val op ofs [proc write-byte])
  (io-at op ofs (λ (op) (write-pascal-bytes val op proc))))

;; read-pascal-string : input-port [procedure] -> string
;; read the length using proc then read length char from input port ip
(define (read-pascal-string ip [proc read-byte])
  (read-string (proc ip) ip))
;(printf "~a\n" (read-pascal-string (open-input-string "\x05NeigeABCDEF")))

;; write-pascal-string : string output-port [procedure] -> any
;; write the length of the string val using proc then write the string val to output port op
(define (write-pascal-string val op [proc write-byte])
  (proc (string-length val) op)
  (write-string val op))
#;
(let ([osp (open-output-bytes)])
  (write-pascal-string "Neige" osp)
  (get-output-bytes osp))

;; read-pascal-string-at : input-port integer [procedure] -> string
;; read the length using proc then read length char from offset ofs of input port ip
(define (read-pascal-string-at ip ofs [proc read-byte])
  (io-at ip ofs (λ (ip) (read-pascal-string ip proc))))

;; write-pascal-string-at : string output-port integer [procedure] -> any
;; write the length of the string val using proc then write the string val to offset ofs of output port op
(define (write-pascal-string-at val op ofs [proc write-byte])
  (io-at op ofs (λ (op) (write-pascal-string val op proc))))
