; The humble pope Neige the reversing schemer and avatar of lazyness
; officially declares this memory mapping library
; to be (K) 3176 YOLD
; Do what thou wilt with thine copy

#lang scheme/base

(require scheme/contract)

(provide/contract [default-mapper (-> (-> natural-number/c natural-number/c))]
                  [set-default-mapper! (-> (-> natural-number/c natural-number/c) any)]
                  [linear-mapper (-> natural-number/c natural-number/c)]
                  [snes-lorom-mapper (-> natural-number/c natural-number/c)]
                  [snes-hirom-mapper (-> natural-number/c natural-number/c)]
                  [map-position (->* (port? natural-number/c)
                                     ((-> natural-number/c natural-number/c))
                                     any)])

;; A memory-mapper is a procedure returning an
;; integer (PC file address) when applied to an integer

;; Consumes a file address and returns a file address
(define (linear-mapper a) a)

;; Consumes a SNES ROM CPU address and returns a file address
(define (snes-lorom-mapper a)
  (let ((bank (arithmetic-shift (bitwise-and a #x7F0000) -16)) ; bank #x80-#xFF are mirror of bank #x00-#x7F
        (word (bitwise-and a #xFFFF)))
    (cond
      [(and (<= #x00 bank #x3F)
            (<= #x8000 word #xFFFF))
       (+ (* bank #x8000) (- word #x8000))]
      [(<= #x40 bank #x7C)
       (if (<= #x0000 word #x7FFF)
           (+ (* bank #x8000) word)
           (+ (* bank #x8000) (- word #x8000)))]
      [else (error 'snes-lorom-mapper "not a ROM address #x~x" a)])))

;; Consumes a SNES ROM CPU address and returns a file address
(define (snes-hirom-mapper a)
  (let ((bank (arithmetic-shift (bitwise-and a #xFF0000) -16))
        (word (bitwise-and a #xFFFF)))
    (cond
      [(and (<= #x00 bank #x3F)
            (<= #x8000 word #xFFFF))
       (+ (* bank #x10000) word)]
      [(<= #x40 bank #x7D)
       (+ (* (- bank #x40) #x10000) word)]
      [(and (<= #x80 bank #xBF)
            (<= #x8000 word #xFFFF))
       (+ (* (- bank #x80) #x10000) word)]
      [(<= #xC0 bank #xFE)
       (+ (* (- bank #x80 #x40) #x10000) word)]
      [(= bank #xFF)
       (+ (* (- bank #x80 #x40) #x10000) word)]
      [else (error 'snes-hirom-mapper "not a ROM address #x~x" a)])))

;; default-mapper :  -> procedure
;; return the default memory-mapper
;; set-default-mapper! : procedure -> void
;; sets the default memory-mapper
(define-values (default-mapper
                set-default-mapper!)
  (let ([current-mapper linear-mapper])
    (values
     (lambda ()
       current-mapper)
     (lambda (mapper)
       (set! current-mapper mapper)))))

;; map-position : port integer [memory-mapper] -> void
(define (map-position p ofs [mapper (default-mapper)])
  (file-position p (mapper ofs)))
