; The humble pope Neige the reversing schemer and avatar of lazyness
; officially declares this free space manager library
; to be (K) 3176 YOLD
; Do what thou wilt with thine copy

#lang scheme

(provide make-free-space-handler)

;; commands
;;   add : integer integer -> void
;;     Consumes an address and a size and add them to the list
;;   add-list : integer integer -> void
;;     Consumes a list of list containing an address and a size and add them to the list
;;   request-anywhere : integer -> integer or #f
;;     Consumes a size and return an address or #f if no space available
;;   request-in-range : integer integer integer -> integer or #f
;;     Consumes a start address, a end address and a size and
;;     return an address or #f if no space available
;;   request-at : integer integer -> integer or #f
;;     Consumes an address and a size and
;;     return the address or #f if no space available
;;   update : integer integer -> void
;;     Consumes an address and the size consumed and update the list
;;   dump :  -> assoc list
(define (make-free-space-handler)
  (letrec ([free-space '()]
           [find (λ (pred sel lst col)
                   (if (null? lst)
                       (col '() '())
                       (if (pred (sel (car lst)))
                           (col '() lst)
                           (find pred
                                 sel
                                 (cdr lst)
                                 (λ (seen rest)
                                   (col (cons (car lst) seen) rest))))))]
           [lass= (λ (addr)
                    (find (λ (x) (= addr x))
                          car
                          free-space
                          (λ (seen rest)
                            (if (null? rest) #f (car rest)))))]
           [rass= (λ (size)
                    (find (λ (x) (= size x))
                          cdr
                          free-space
                          (λ (seen rest)
                            (if (null? rest) #f (car rest)))))]
           [insert (λ (addr size)
                     (let-values ([(seen rest) (find (λ (x) (< size x)) cdr free-space values)])
                       (set! free-space (append seen (list (cons addr size)) rest))))]
           [remove (λ (addr)
                     (let-values ([(seen rest) (find (λ (x) (= addr x)) car free-space values)])
                       (set! free-space (append seen (cdr rest)))))])
    (match-lambda*
      [(list 'add (? integer? off) (? integer? siz)) (insert off siz)]
      [(list 'add-list (and loc (list (list (? integer?) (? integer?)) ...))) (for-each (λ (p) (insert (car p) (cadr p))) loc)]
      [(list 'request-anywhere (? integer? size))
       (find (λ (x) (<= size x))
             cdr
             free-space
             (λ (seen rest)
               (if (null? rest) #f (caar rest))))]
      [(list 'request-in-range (? integer? range-lo) (? integer? range-hi) (? integer? size))
       (let loop ([found (find (λ (x) (<= range-lo x range-hi))
                               car
                               free-space
                               (λ (seen rest)
                                 rest))])
         (if (null? found)
             #f
             (let* ([found-addr (caar found)]
                    [found-size (cdar found)]
                    [in-range-size (if (< (+ found-addr found-size) range-hi)
                                       found-size
                                       (- range-hi found-addr))])
               (if (>= in-range-size size)
                   found-addr
                   (loop (find (λ (x) (<= range-lo x range-hi))
                               car
                               (cdr found)
                               (λ (seen rest)
                                 rest)))))))]
      [(list 'request-at (? integer? addr) (? integer? size))
       (let loop ([found (find (λ (x) (<= x addr))
                               car
                               free-space
                               (λ (seen rest)
                                 rest))])
         (if (null? found)
             #f
             (let* ([found-addr (caar found)]
                    [found-size (cdar found)]
                    #;[in-range-size (if (< (+ found-addr found-size) range-hi)
                                       found-size
                                       (- range-hi found-addr))])
               (if (and (<= addr (+ found-addr found-size))
                        (<= size (- (+ found-addr found-size) addr)))
                   (begin
                     (unless (= addr found-addr) ; split the free space so that update remove the correct range
                       (remove found-addr)
                       (insert found-addr (- addr found-addr))
                       (insert addr (- found-size (- addr found-addr))))
                     addr)
                   (loop (find (λ (x) (<= x addr))
                               car
                               (cdr found)
                               (λ (seen rest)
                                 rest)))))))]
      [(list 'update (? integer? addr) (? integer? space-consumed))
       (let ([orig-size (cdr (lass= addr))])
         (remove addr)
         (insert (+ addr space-consumed) (- orig-size space-consumed)))]
      [(list 'dump) free-space])))
#;
(begin
  (define fsh (make-free-space-handler))
  (fsh 'add #x03E79E #x7CFD)
  (fsh 'add #x04BDB5 #x3F66)
  (fsh 'add #x147300 #x1908)
  (fsh 'add #x17D616 #x08F2)
  (fsh 'add #x17E3A1 #x006D)
  (for-each (λ (p) (printf "#x~x #x~x\n" (car p) (cdr p))) (fsh 'dump))(newline)
  (printf "#x~x\n\n" (fsh 'request-at #x17D616 #x0050))
  (for-each (λ (p) (printf "#x~x #x~x\n" (car p) (cdr p))) (fsh 'dump))(newline)
  (printf "#x~x\n\n" (fsh 'request-at #x17D716 #x0050))
  (for-each (λ (p) (printf "#x~x #x~x\n" (car p) (cdr p))) (fsh 'dump))(newline)
  ;(printf "#x~x\n" (fsh 'request-in-range #x170000 #x17FFFF #x0050))
  ;(printf "#x~x\n" (fsh 'request-in-range #x170000 #x17FFFF #x0500))
  ;(printf "#x~x\n" (fsh 'request-in-range #x030000 #x03FFFF #x1000))
  ;(printf "#x~x\n" (fsh 'request-in-range #x030000 #x03FFFF #x2000))
  )
