; The humble pope Neige the reversing schemer and avatar of lazyness
; officially declares this byte conversion table library
; to be (K) 3176 YOLD
; Do what thou wilt with thine copy

#lang scheme/base

(require scheme/match
         scheme/contract)

(provide/contract [make-table (-> (symbols 'extraction 'insertion) any)])

; standard table entry line
; endline "*FE" or "*FE=<End Text>"
; endstring "/FF=<end>" or "/<end>"
; endsection "\FD"
; table entry "hex=string"
; linked entry "$0500=<Color>,1"
; dakuten control code "!F0"
; handakuten control code "@F1"
; bookmark "(8000h)Text1"
; script dump entry "[8000h-8450h]Block 1"
; script insert bookmark "{8000h-TextDump.txt}Block-e 1"

(define (make-table mode)
  (let* ([table (make-hash)]
         [max-hex-length 0]
         [max-text-length (make-vector 256 0)]
         [add-mapping
          (λ (k v)
            (let ([hex-len (bytes-length k)]
                  [text-len (if (bytes? v) (bytes-length v) 0)]
                  [text-index (if (bytes? v) (bytes-ref v 0) 0)])
              (when (> hex-len max-hex-length)
                (set! max-hex-length hex-len))
              (when (> text-len (vector-ref max-text-length text-index))
                (vector-set! max-text-length text-index text-len)))
            (if (eq? mode 'extraction)
                (hash-set! table k v)
                (hash-set! table v k)))]
         [add-mappings
          (λ (lom)
            (for-each (λ (c) (add-mapping (car c) (cadr c))) lom))]
         [number->bytes
          (λ (n)
            (let ([numstr (string->bytes/latin-1 (string-upcase (number->string n 16)))])
              (bytes-append #"<$" (if (= (bytes-length numstr) 1) #"0" #"") numstr #">")))]
         [add-default-insert-mappings
          (λ ()
            (for ([i (in-range 256)])
              (add-mapping (bytes i) (number->bytes i))))]
         [read-table-file
          (λ (fname)
            (call-with-input-file fname
              (λ (ip)
                (letrec ([bytes->numbers
                          (λ (b)
                            (if (< (bytes-length b) 2)
                                '()
                                (cons (string->number (bytes->string/latin-1 (subbytes b 0 2)) 16)
                                      (bytes->numbers (subbytes b 2)))))]
                         [table-pair-regexp #px#"^((?:[0-9A-Fa-f]{2})+?)[ =](.*?)\n"])
                  (let loop ([p (regexp-match table-pair-regexp ip)])
                    (if p
                        (cons (list (list->bytes (bytes->numbers (cadr p))) (caddr p))
                              (loop (regexp-match table-pair-regexp ip)))
                        '()))))
              #:mode 'text))])
    (when (eq? mode 'insertion)
      (add-default-insert-mappings))
    (match-lambda*
      [(list 'get-mode) mode]
      [(list 'get-max-length) max-hex-length]
      [(list 'get-max-length (? byte? b)) (vector-ref max-text-length b)]
      [(list 'get-max-length (? bytes? b)) (vector-ref max-text-length (bytes-ref b 0))]
      [(list 'add-mapping (? bytes? k) (? (λ (x) (or (bytes? x) (procedure? x))) v)) (add-mapping k v)]
      [(list 'add-mappings (and lom (list (list (? bytes?) (? (λ (x) (or (bytes? x) (procedure? x))))) ...))) (add-mappings lom)]
      [(list 'get-mapping k ...) (apply hash-ref table k)]
      [(list 'read-table (? path-string? f)) (add-mappings (read-table-file f))]
      [(list 'dump) table])))

#;(begin
  (define test-table-e (make-table 'extraction))
  (test-table-e 'read-table "introtable-iso8859-1.tbl")
  (test-table-e 'get-mode)
  (test-table-e 'add-mapping #"\x01" #"test1")
  (test-table-e 'get-mapping #"\x01")
  (test-table-e 'get-mapping #"\x02" #"<$02>")
  (test-table-e 'add-mappings '((#"\x03" #"test3") (#"\x04" #"test4")))
  (test-table-e 'get-mapping #"\x03")
  (test-table-e 'get-mapping #"\x04")
  (test-table-e 'get-max-length))

#;(begin
  (define test-table-i (make-table 'insertion))
  (test-table-i 'read-table "introtable-iso8859-1.tbl")
  (test-table-i 'get-mode)
  (test-table-i 'add-mapping #"\x01" #"test1")
  (test-table-i 'get-mapping #"test1")
  (test-table-i 'get-mapping #"<$02>" (λ () #"\x02"))
  (test-table-i 'add-mappings '((#"\x03" #"test3") (#"\x04" #"test4")))
  (test-table-i 'get-mapping #"test3")
  (test-table-i 'get-mapping #"test4")
  (test-table-i 'get-max-length #x74)
  (test-table-i 'get-max-length #x20))
